import sys 
class ChatLLM:
    def __init__(self,model,tokenizer,
                 model_type=None,
                 max_chat_rounds=20,
                 max_new_tokens=512,
                 stream=True,
                 history=None
                ):
        self.model = model
        self.tokenizer = tokenizer
        self.model.generation_config.max_new_tokens = max_new_tokens
        self.model.eval()
        self.history = [] if history is None else history
        self.max_chat_rounds = max_chat_rounds
        self.stream = stream
        
        try:
            self.register_magic() 
            response = self('你好')
            print(response)
            print('register magic %%chat sucessed ...',file=sys.stderr)
            self.history = self.history[:-1]
        except Exception as err:
            print('register magic %%chat failed ...',file=sys.stderr)
            print(err, file=sys.stderr)
            
            
    @classmethod
    def build_messages(cls,query=None,history=None,system=None):
        messages = []
        history = history if history else [] 
        if system is not None:
            messages.append({'role':'system','content':system})
        for prompt,response in history:
            pair = [{"role": "user", "content": prompt},
                {"role": "assistant", "content": response}]
            messages.extend(pair)
        if query is not None:
            messages.append({"role": "user", "content": query})
        return messages


        
    def chat(self, messages, stream=True):
        
        model,tokenizer = self.model,self.tokenizer
        text = tokenizer.apply_chat_template(
            messages,
            tokenize=False,
            add_generation_prompt=True
        )
        model_inputs = tokenizer([text], return_tensors="pt").to(model.device)
        
        if stream:
            from transformers import TextStreamer
            streamer = TextStreamer(tokenizer, skip_prompt=True, skip_special_tokens=True)
        else:
            streamer = None

        generated_ids = model.generate(
            **model_inputs,
            streamer  = streamer 
        )

        generated_ids = [
            output_ids[len(input_ids):] for input_ids, 
            output_ids in zip(model_inputs.input_ids, generated_ids)
        ]

        response = tokenizer.batch_decode(generated_ids, skip_special_tokens=True)[0]
        return response 
        

    def __call__(self,query):
        
        len_his = len(self.history)
        if len_his>=self.max_chat_rounds+1:
            self.history = self.history[len_his-self.max_chat_rounds:]
        messages = self.build_messages(query=query,history=self.history)
        response = self.chat(messages,stream=self.stream)
        self.history.append((query,response))
        
        if self.stream:
            from IPython.display import clear_output,display
            clear_output(wait=True) 
        return response
    
    
    def register_magic(self):
        import IPython
        from IPython.core.magic import (Magics, magics_class, line_magic,
                                        cell_magic, line_cell_magic)
        @magics_class
        class ChatMagics(Magics):
            def __init__(self,shell, pipe):
                super().__init__(shell)
                self.pipe = pipe

            @line_cell_magic
            def chat(self, line, cell=None):
                "Magic that works both as %chat and as %%chat"
                if cell is None:
                    return self.pipe(line)
                else:
                    print(self.pipe(cell))       
        ipython = IPython.get_ipython()
        magic = ChatMagics(ipython,self)
        ipython.register_magics(magic)
        


# Pytorch Tabular
# Author: Manu Joseph <manujoseph@gmail.com>
# For license information, see LICENSE.TXT
import torch
import torch.nn as nn

from ..common.heads import blocks
from ..common.layers import Add, Embedding1dLayer, GatedFeatureLearningUnit, NeuralDecisionTree
from ..common.layers.activations import entmax15, entmoid15, sparsemax, sparsemoid, t_softmax

from ..base_model import BaseModel

class GatedAdditiveTreesBackbone(nn.Module):
    ACTIVATION_MAP = {
        "entmax": entmax15,
        "sparsemax": sparsemax,
        "softmax": nn.functional.softmax,
        "t-softmax": t_softmax,
    }

    BINARY_ACTIVATION_MAP = {
        "entmoid": entmoid15,
        "sparsemoid": sparsemoid,
        "sigmoid": nn.functional.sigmoid,
    }

    def __init__(
        self,
        cat_embedding_dims: list,
        n_continuous_features: int,
        gflu_stages: int,
        num_trees: int,
        tree_depth: int,
        chain_trees: bool = True,
        tree_wise_attention: bool = False,
        tree_wise_attention_dropout: float = 0.0,
        gflu_dropout: float = 0.0,
        tree_dropout: float = 0.0,
        binning_activation: str = "entmoid",
        feature_mask_function: str = "softmax",
        gflu_feature_init_sparsity: float = 0.3,
        tree_feature_init_sparsity: float = 0.8,
        learnable_sparsity: bool = True,
        batch_norm_continuous_input: bool = True,
        virtual_batch_size: int = None,
        embedding_dropout: float = 0.0,
    ):
        super().__init__()
        assert (
            binning_activation in self.BINARY_ACTIVATION_MAP.keys()
        ), f"`binning_activation should be one of {self.BINARY_ACTIVATION_MAP.keys()}"
        assert (
            feature_mask_function in self.ACTIVATION_MAP.keys()
        ), f"`feature_mask_function should be one of {self.ACTIVATION_MAP.keys()}"

        self.gflu_stages = gflu_stages
        self.num_trees = num_trees
        self.tree_depth = tree_depth
        self.chain_trees = chain_trees
        self.tree_wise_attention = tree_wise_attention
        self.tree_wise_attention_dropout = tree_wise_attention_dropout
        self.gflu_dropout = gflu_dropout
        self.tree_dropout = tree_dropout
        self.binning_activation = self.BINARY_ACTIVATION_MAP[binning_activation]
        self.feature_mask_function = self.ACTIVATION_MAP[feature_mask_function]
        self.batch_norm_continuous_input = batch_norm_continuous_input
        self.n_continuous_features = n_continuous_features
        self.cat_embedding_dims = cat_embedding_dims
        self._embedded_cat_features = sum([y for x, y in cat_embedding_dims])
        self.n_features = self._embedded_cat_features + n_continuous_features
        self.embedding_dropout = embedding_dropout
        self.output_dim = 2**self.tree_depth if self.num_trees > 0 else self.n_features
        self.gflu_feature_init_sparsity = gflu_feature_init_sparsity
        self.tree_feature_init_sparsity = tree_feature_init_sparsity
        self.learnable_sparsity = learnable_sparsity
        self.virtual_batch_size = virtual_batch_size
        self._build_network()

    def _build_network(self):
        if self.gflu_stages > 0:
            self.gflus = GatedFeatureLearningUnit(
                n_features_in=self.n_features,
                n_stages=self.gflu_stages,
                feature_mask_function=self.feature_mask_function,
                dropout=self.gflu_dropout,
                feature_sparsity=self.gflu_feature_init_sparsity,
                learnable_sparsity=self.learnable_sparsity,
            )
        if self.num_trees > 0:
            self.trees = nn.ModuleList(
                [
                    NeuralDecisionTree(
                        depth=self.tree_depth,
                        n_features=self.n_features + 2**self.tree_depth * t if self.chain_trees else self.n_features,
                        dropout=self.tree_dropout,
                        binning_activation=self.binning_activation,
                        feature_mask_function=self.feature_mask_function,
                        feature_sparsity=self.tree_feature_init_sparsity,
                        learnable_sparsity=self.learnable_sparsity,
                    )
                    for t in range(self.num_trees)
                ]
            )
            if self.tree_wise_attention:
                self.tree_attention = nn.MultiheadAttention(
                    embed_dim=self.output_dim,
                    num_heads=1,
                    batch_first=False,
                    dropout=self.tree_wise_attention_dropout,
                )

    def _build_embedding_layer(self):
        return Embedding1dLayer(
            continuous_dim=self.n_continuous_features,
            categorical_embedding_dims=self.cat_embedding_dims,
            embedding_dropout=self.embedding_dropout,
            batch_norm_continuous_input=self.batch_norm_continuous_input,
            virtual_batch_size=self.virtual_batch_size,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.gflu_stages > 0:
            x = self.gflus(x)
        if self.num_trees > 0:
            # Decision Tree
            tree_outputs = []
            tree_feature_masks = []  # TODO make this optional and create feat importance
            tree_input = x
            for i in range(self.num_trees):
                tree_output, feat_masks = self.trees[i](tree_input)
                tree_outputs.append(tree_output.unsqueeze(-1))
                tree_feature_masks.append(feat_masks)
                if self.chain_trees:
                    tree_input = torch.cat([tree_input, tree_output], 1)
            tree_outputs = torch.cat(tree_outputs, dim=-1)
            if self.tree_wise_attention:
                tree_outputs = tree_outputs.permute(2, 0, 1)
                tree_outputs, _ = self.tree_attention(tree_outputs, tree_outputs, tree_outputs)
                tree_outputs = tree_outputs.permute(1, 2, 0)
            return tree_outputs
        else:
            return x

    @property
    def feature_importance_(self):
        return self.gflus.feature_mask_function(self.gflus.feature_masks).sum(dim=0).detach().cpu().numpy()


class CustomHead(nn.Module):
    """Custom Head for GATE.

    """

    def __init__(self, input_dim: int, hparams):
        super().__init__()
        self.hparams = hparams
        self.input_dim = input_dim
        if self.hparams.share_head_weights:
            self.head = self._get_head_from_config()
        else:
            self.head = nn.ModuleList([self._get_head_from_config() for _ in range(self.hparams.num_trees)])
        # random parameter with num_trees elements
        self.eta = nn.Parameter(torch.rand(self.hparams.num_trees, requires_grad=True))
        if self.hparams.task == "regression":
            self.T0 = nn.Parameter(torch.rand(self.hparams.output_dim), requires_grad=True)

    def _get_head_from_config(self):
        _head_callable = getattr(blocks, self.hparams.head)
        return _head_callable(
            in_units=self.input_dim,
            output_dim=self.hparams.output_dim,
            config=_head_callable._config_template(**self.hparams.head_config),
        )  # output_dim auto-calculated from other configs

    def forward(self, backbone_features: torch.Tensor) -> torch.Tensor:
        # B x L x T
        if not self.hparams.share_head_weights:
            # B x T X Output
            y_hat = torch.cat(
                [h(backbone_features[:, :, i]).unsqueeze(1) for i, h in enumerate(self.head)],
                dim=1,
            )
        else:
            # https://discuss.pytorch.org/t/how-to-pass-a-3d-tensor-to-linear-layer/908/6
            # B x T x L -> B x T x Output
            y_hat = self.head(backbone_features.transpose(2, 1))

        # applying weights to each tree and summing up
        # ETA
        y_hat = y_hat * self.eta.reshape(1, -1, 1)
        # summing up
        y_hat = y_hat.sum(dim=1)

        if self.hparams.task == "regression":
            y_hat = y_hat + self.T0
        return y_hat


class GatedAdditiveTreeEnsembleModel(BaseModel):
    def __init__(self, config, **kwargs):
        super().__init__(config, **kwargs)

    @property
    def backbone(self):
        return self._backbone

    @property
    def embedding_layer(self):
        return self._embedding_layer

    @property
    def head(self):
        return self._head

    def _build_network(self):
        # Backbone
        self._backbone = GatedAdditiveTreesBackbone(
            n_continuous_features=self.hparams.continuous_dim,
            cat_embedding_dims=self.hparams.embedding_dims,
            gflu_stages=self.hparams.gflu_stages,
            gflu_dropout=self.hparams.gflu_dropout,
            num_trees=self.hparams.num_trees,
            tree_depth=self.hparams.tree_depth,
            tree_dropout=self.hparams.tree_dropout,
            binning_activation=self.hparams.binning_activation,
            feature_mask_function=self.hparams.feature_mask_function,
            batch_norm_continuous_input=self.hparams.batch_norm_continuous_input,
            chain_trees=self.hparams.chain_trees,
            tree_wise_attention=self.hparams.tree_wise_attention,
            tree_wise_attention_dropout=self.hparams.tree_wise_attention_dropout,
            gflu_feature_init_sparsity=self.hparams.gflu_feature_init_sparsity,
            tree_feature_init_sparsity=self.hparams.tree_feature_init_sparsity,
            virtual_batch_size=self.hparams.virtual_batch_size,
        )
        # Embedding Layer
        self._embedding_layer = self._backbone._build_embedding_layer()
        # Head
        if self.hparams.num_trees == 0:
            self.T0 = nn.Parameter(torch.rand(self.hparams.output_dim), requires_grad=True)
            self._head = nn.Sequential(self._get_head_from_config(), Add(self.T0))
        else:
            self._head = CustomHead(self.backbone.output_dim, self.hparams)

    def data_aware_initialization(self, dataloader):
        if self.hparams.task == "regression":
            batch = next(iter(dataloader))
            t0 = torch.mean(batch["target"], dim=0)
            if self.hparams.num_trees != 0:
                self.head.T0.data = t0
            else:
                self.T0.data = t0



import numpy as np
import pandas as pd 
import datetime 
from tqdm.auto  import tqdm 

def printlog(info: str) -> None:
    now_time = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
    print("\n" + "==========" * 8 + "%s" % now_time)
    print(info + '...\n\n')

def prettydf(df):
    from tabulate import tabulate
    return tabulate(df,headers=df.columns, tablefmt="pretty")

def process_data(df, yhat_col = 'yhat', 
                 y_col = 'y', 
                 id_col = 'id', 
                 model_col = 'model'):
    for col in [yhat_col,id_col,model_col]:
        assert col in df.columns

    dfdata = df.pivot_table(index=  id_col, 
                                 columns = model_col, 
                                 values = yhat_col)
    dfdata.columns.name = None
    if y_col in df.columns:
        dflabels = df.pivot_table(index = id_col, columns = model_col, values = y_col)
        assert np.all(dflabels.max(axis=1) == dflabels.min(axis=1))
        labels = dflabels.mean(axis=1)
        assert np.all(dfdata.index==labels.index)
        dfdata[y_col] = labels 
    return dfdata 

class Stacker:
    def __init__(self, model_names, 
                 score_fn, 
                 greater_is_better = True
                ):
        
        self.model_names = model_names
        self.score_fn = score_fn 
        self.greater_is_better = greater_is_better 
        
    def fit_greedy_stacking(self, dfstack, num_models = None, num_warmup = None):
        
        assert 'y' in dfstack.columns, 'dfstack should with a label column  y'

        for name in self.model_names:
            assert name in dfstack.columns, f'{name} not in dfstack.columns.'
    
        if num_models is None:
            num_models = len(self.model_names)*200 
        if num_warmup is None:
            num_warmup = len(self.model_names)*5

        dfstack = dfstack.copy()
        
        printlog('step1: sort base models by score ...')
        
        scores = [self.score_fn(dfstack['y'],dfstack[name]) for name in self.model_names]
        dfscores = pd.DataFrame({'model': self.model_names, 'score': scores})
        dfscores = dfscores.sort_values('score',
                    ascending= False if self.greater_is_better else True).copy()
        
        dfscores.index = range(len(dfscores))
        print(prettydf(dfscores))
        
        
        dfscores['weights'] = [0 for _ in dfscores['model']]
        sorted_models = dfscores['model'].tolist()
        dfstack = dfstack[sorted_models+['y']]
        
        loop = tqdm(range(num_models),total = num_models)
        
        printlog('step2: start greedy stacking...')
        
        for i in loop:
            
            if i==0:
                cur_yhat = dfstack[sorted_models[0]].tolist()
                cur_score = dfscores['score'].iloc[0]
                dfscores.loc[0,'weights'] = 1
                loop.set_postfix(**{'i':i,'score':cur_score})
        
            elif i+1<=num_warmup:
                k = i%(len(self.model_names))
                dfscores.loc[k,'weights'] = 1 + dfscores['weights'].loc[k]
                cur_yhat = (dfstack[sorted_models].values)@(dfscores['weights'].values)/dfscores['weights'].sum()
                cur_score = self.score_fn(dfstack['y'],cur_yhat)
                loop.set_postfix(**{'i':i,'score':cur_score})
            else:
                for j in range(len(self.model_names)):
                    maybe_weight = dfscores['weights'].tolist()
                    maybe_weight[j] = maybe_weight[j] + 1 
                    maybe_weight = np.array(maybe_weight)
                    maybe_yhat = (dfstack[sorted_models].values)@maybe_weight/maybe_weight.sum()
                    maybe_score = self.score_fn(dfstack['y'],maybe_yhat)
                    
                    if (self.greater_is_better and maybe_score>=cur_score) or (
                        self.greater_is_better==False and maybe_score<=cur_score):
                        cur_yhat = maybe_yhat
                        cur_score = maybe_score
                        dfscores.loc[j,'weights'] = dfscores['weights'][j]+1
                        loop.set_postfix(**{'i':i,'score':cur_score})
                        break
                else:
                    loop.set_postfix(**{'i':i,'score':cur_score})
                    loop.close()
                    print('could not get better score, early stopping...')
                    break 
        
        printlog('step3: save result...')
        dfstack['yhat'] = cur_yhat
        print('best_score = ', cur_score)
        
        total_weights = dfscores['weights'].sum()
        dfensemble = pd.DataFrame({'model':['GreedyStacking'],
                                 'score':[cur_score],'weights':[total_weights]})
        
        dfsummary = pd.concat([dfscores,dfensemble],ignore_index=True)
        dfsummary['weight_ratio'] = np.round(dfsummary['weights']/total_weights,3)
        
        print(prettydf(dfsummary))
        
        self.dfscores = dfscores
        self.dfsummary = dfsummary
        self.sorted_models = sorted_models
        return dfstack 

    def fit_optuna_stacking(self, dfstack, n_trials=500, timeout=1200):
    
        import optuna
        optuna.logging.set_verbosity(optuna.logging.WARNING)
        dfstack = dfstack.copy()
        
        printlog('step1: sort base models by score ...')
        scores = [self.score_fn(dfstack['y'],dfstack[name]) for name in self.model_names]
        dfscores = pd.DataFrame({'model': self.model_names, 'score': scores})
        dfscores = dfscores.sort_values('score',
                    ascending= False if self.greater_is_better else True) 
        dfscores.index = range(len(dfscores))
        sorted_models = dfscores['model'].tolist()
        print(prettydf(dfscores))
        
        
        def objective(trial):
            weights_dict = {name: trial.suggest_int(name, 1, 100) for name in sorted_models}
            weights = np.array([weights_dict[name] for name in sorted_models])
            yhat = (dfstack[sorted_models].values)@weights/(weights.sum())
            score = self.score_fn(dfstack['y'], yhat)
            return score
    
        study = optuna.create_study(
            direction="maximize" if self.greater_is_better else "minimize",
            study_name="optuna_stacking"
        )
    
        printlog('step2: start optuna search ...')
        study.optimize(objective, n_trials=n_trials, timeout=timeout, show_progress_bar=True)
    
        printlog('step3: save result...')
        best_weights = study.best_params
        best_score = study.best_value
        
        weights = np.array([best_weights[name] for name in sorted_models])
        total_weights = weights.sum()
        stack_yhat = (dfstack[sorted_models].values)@weights/(weights.sum())
    
        dfstack['yhat'] = stack_yhat 
        print('best_score = ', best_score)
    
        dfscores['weights'] = weights
        
        dfensemble = pd.DataFrame({'model':['OptunaStacking'],'score':[best_score],'weights':[total_weights]})
        dfsummary = pd.concat([dfscores,dfensemble],ignore_index=True)
        dfsummary['weight_ratio'] = np.round(dfsummary['weights']/total_weights,3)
        
        print(prettydf(dfsummary))

        self.dfscores = dfscores
        self.dfsummary = dfsummary
        self.sorted_models = sorted_models
        return dfstack 
        
    
    def predict(self,dftest):
        dftest = dftest.copy()
        test_yhat = (dftest[self.sorted_models].values)@(self.dfscores['weights'].values
                    )/self.dfscores['weights'].sum()
        dftest['yhat'] = test_yhat
        if 'y' in dftest.columns:
            test_score = self.score_fn(dftest['y'],test_yhat)
            print('test_score = ', test_score)
        return dftest
